@@ -33,7 +33,6 @@
 #include <QtQuick3D/private/qquick3dobject_p.h>
 
 #include <QtQuick3DUtils/private/qssgutils_p.h>
-#include <QtQuick3DRuntimeRender/private/qssgrenderparticles_p.h>
 
 QT_BEGIN_NAMESPACE
 
@@ -42,7 +41,7 @@ QT_BEGIN_NAMESPACE
     \inherits Particle3D
     \inqmlmodule QtQuick3D.Particles3D
     \brief Particle using a 2D sprite texture.
-    \since 6.1
+    \since 6.2
 
     The SpriteParticle3D is a logical particle element that creates particles
     from a 2D sprite texture.
@@ -55,23 +54,32 @@ QQuick3DParticleSpriteParticle::QQuick3DParticleSpriteParticle(QQuick3DNode *par
         handleMaxAmountChanged(m_maxAmount);
     }));
     m_connections.insert("system", QObject::connect(this, &QQuick3DParticle::systemChanged, [this]() {
-       handleSystemChanged(system());
+        handleSystemChanged(system());
+    }));
+    m_connections.insert("sortMode", QObject::connect(this, &QQuick3DParticle::sortModeChanged, [this]() {
+        markNodesDirty();
     }));
 }
 
 QQuick3DParticleSpriteParticle::~QQuick3DParticleSpriteParticle()
 {
+    if (m_spriteSequence)
+        m_spriteSequence->m_parentParticle = nullptr;
     for (const auto &connection : qAsConst(m_connections))
         QObject::disconnect(connection);
     deleteNodes();
+
+    auto lightList = lights();
+    qmlClearLights(&lightList);
 }
 
 void QQuick3DParticleSpriteParticle::deleteNodes()
 {
-    for (PerEmitterData &value : m_perEmitterData) {
+    for (const PerEmitterData &value : qAsConst(m_perEmitterData)) {
         value.particleUpdateNode->m_spriteParticle = nullptr;
         delete value.particleUpdateNode;
     }
+    m_perEmitterData.clear();
 }
 
 /*!
@@ -122,40 +130,16 @@ QQuick3DTexture *QQuick3DParticleSpriteParticle::sprite() const
 }
 
 /*!
-    \qmlproperty int SpriteParticle3D::frameCount
+    \qmlproperty SpriteSequence3D SpriteParticle3D::spriteSequence
 
-    This property defines the amount of image frames in \l sprite. Particle animates
-    through these frames during its \l {ParticleEmitter3D::lifeSpan}{lifeSpan}.
-    The frames should be laid out horizontally in the same image file. For example,
-    \l sprite could be a \c {512x64} image, with \c frameCount of \c 8. If the particle
-    \l {ParticleEmitter3D::lifeSpan}{lifeSpan} is \c 2000, each of those images will be
-    visible for \c 250 milliseconds.
-
-    The default value is \c 1.
-
-    \sa interpolate
+    This property defines the sprite sequence properties for the particle.
+    If the \l sprite texture contains a frame sequence, set this property
+    to define the frame count, animation direction etc. features.
 */
-int QQuick3DParticleSpriteParticle::frameCount() const
+
+QQuick3DParticleSpriteSequence *QQuick3DParticleSpriteParticle::spriteSequence() const
 {
-    return m_frameCount;
-}
-
-/*!
-    \qmlproperty bool SpriteParticle3D::interpolate
-
-    This property defines if the sprites are interpolated (blended) between frames
-    to make the animation appear smoother.
-
-    \note This property doesn't have an effect when the sprite has only a single
-    frame (so \l frameCount is \c 1).
-
-    The default value is \c true.
-
-    \sa frameCount
-*/
-bool QQuick3DParticleSpriteParticle::interpolate() const
-{
-    return m_interpolate;
+    return m_spriteSequence;
 }
 
 /*!
@@ -206,6 +190,48 @@ QQuick3DTexture *QQuick3DParticleSpriteParticle::colorTable() const
     return m_colorTable;
 }
 
+/*!
+    \qmlproperty list<Light> SpriteParticle3D::lights
+    \since 6.3
+
+    This property contains a list of \l [QtQuick3D QML] {Light}{lights} used
+    for rendering the particles.
+    \note For optimal performance, define lights only if they are needed and keep
+    the amount of lights at minimum.
+*/
+
+QQmlListProperty<QQuick3DAbstractLight> QQuick3DParticleSpriteParticle::lights()
+{
+    return QQmlListProperty<QQuick3DAbstractLight>(this,
+                                            nullptr,
+                                            QQuick3DParticleSpriteParticle::qmlAppendLight,
+                                            QQuick3DParticleSpriteParticle::qmlLightsCount,
+                                            QQuick3DParticleSpriteParticle::qmlLightAt,
+                                            QQuick3DParticleSpriteParticle::qmlClearLights);
+}
+
+/*!
+    \qmlproperty float QQuick3DParticleSpriteParticle::offsetX
+    \since 6.3
+
+    This property defines the particles offset in the X axis
+*/
+float QQuick3DParticleSpriteParticle::offsetX() const
+{
+    return m_offset.x();
+}
+
+/*!
+    \qmlproperty float QQuick3DParticleSpriteParticle::offsetY
+    \since 6.3
+
+    This property defines the particles offset in the Y axis
+*/
+float QQuick3DParticleSpriteParticle::offsetY() const
+{
+    return m_offset.y();
+}
+
 void QQuick3DParticleSpriteParticle::setBlendMode(BlendMode blendMode)
 {
     if (m_blendMode == blendMode)
@@ -232,22 +258,15 @@ void QQuick3DParticleSpriteParticle::setSprite(QQuick3DTexture *sprite)
     Q_EMIT spriteChanged();
 }
 
-void QQuick3DParticleSpriteParticle::setFrameCount(int frameCount)
+void QQuick3DParticleSpriteParticle::setSpriteSequence(QQuick3DParticleSpriteSequence *spriteSequence)
 {
-    if (m_frameCount == frameCount)
+    if (m_spriteSequence == spriteSequence)
         return;
-    m_frameCount = frameCount;
-    markNodesDirty();
-    Q_EMIT frameCountChanged();
-}
 
-void QQuick3DParticleSpriteParticle::setInterpolate(bool interpolate)
-{
-    if (m_interpolate == interpolate)
-        return;
-    m_interpolate = interpolate;
+    m_spriteSequence = spriteSequence;
+    updateFeatureLevel();
     markNodesDirty();
-    Q_EMIT interpolateChanged();
+    Q_EMIT spriteSequenceChanged();
 }
 
 void QQuick3DParticleSpriteParticle::setBillboard(bool billboard)
@@ -281,10 +300,29 @@ void QQuick3DParticleSpriteParticle::setColorTable(QQuick3DTexture *colorTable)
     });
 
     m_colorTable = colorTable;
+    updateFeatureLevel();
     markNodesDirty();
     Q_EMIT colorTableChanged();
 }
 
+void QQuick3DParticleSpriteParticle::setOffsetX(float value)
+{
+    if (qFuzzyCompare(value, m_offset.x()))
+        return;
+
+    m_offset.setX(value);
+    emit offsetXChanged();
+}
+
+void QQuick3DParticleSpriteParticle::setOffsetY(float value)
+{
+    if (qFuzzyCompare(value, m_offset.y()))
+        return;
+
+    m_offset.setY(value);
+    emit offsetYChanged();
+}
+
 void QQuick3DParticleSpriteParticle::itemChange(QQuick3DObject::ItemChange change,
                                           const QQuick3DObject::ItemChangeData &value)
 {
@@ -307,6 +345,27 @@ static QSSGRenderParticles::BlendMode mapBlendMode(QQuick3DParticleSpriteParticl
     }
 }
 
+QSSGRenderParticles::FeatureLevel QQuick3DParticleSpriteParticle::mapFeatureLevel(QQuick3DParticleSpriteParticle::FeatureLevel level)
+{
+    switch (level) {
+    case QQuick3DParticleSpriteParticle::Simple:
+        return QSSGRenderParticles::FeatureLevel::Simple;
+    case QQuick3DParticleSpriteParticle::Mapped:
+        return QSSGRenderParticles::FeatureLevel::Mapped;
+    case QQuick3DParticleSpriteParticle::Animated:
+        return QSSGRenderParticles::FeatureLevel::Animated;
+    case QQuick3DParticleSpriteParticle::SimpleVLight:
+        return QSSGRenderParticles::FeatureLevel::SimpleVLight;
+    case QQuick3DParticleSpriteParticle::MappedVLight:
+        return QSSGRenderParticles::FeatureLevel::MappedVLight;
+    case QQuick3DParticleSpriteParticle::AnimatedVLight:
+        return QSSGRenderParticles::FeatureLevel::AnimatedVLight;
+    default:
+        Q_ASSERT(false);
+        return QSSGRenderParticles::FeatureLevel::Simple;
+    }
+}
+
 QSSGRenderGraphObject *QQuick3DParticleSpriteParticle::ParticleUpdateNode::updateSpatialNode(QSSGRenderGraphObject *node)
 {
     if (m_spriteParticle) {
@@ -346,7 +405,10 @@ QSSGRenderGraphObject *QQuick3DParticleSpriteParticle::updateParticleNode(const
     auto particles = static_cast<QSSGRenderParticles *>(node);
 
     const auto &perEmitter = perEmitterData(updateNode);
-    updateParticleBuffer(perEmitter, particles);
+    if (m_featureLevel == QQuick3DParticleSpriteParticle::Animated || m_featureLevel == QQuick3DParticleSpriteParticle::AnimatedVLight)
+        updateAnimatedParticleBuffer(perEmitter, particles);
+    else
+        updateParticleBuffer(perEmitter, particles);
 
     if (!updateNode->m_nodeDirty)
         return particles;
@@ -354,25 +416,42 @@ QSSGRenderGraphObject *QQuick3DParticleSpriteParticle::updateParticleNode(const
     if (perEmitter.particleCount == 0)
         return particles;
 
-    if (m_sprite) {
+    if (m_sprite)
         particles->m_sprite = m_sprite->getRenderImage();
-        particles->m_spriteImageCount = m_frameCount;
-        particles->m_blendImages = m_interpolate;
-    } else {
+    else
         particles->m_sprite = nullptr;
+
+    if (m_spriteSequence) {
+        particles->m_spriteImageCount = m_spriteSequence->m_frameCount;
+        particles->m_blendImages = m_spriteSequence->m_interpolate;
+    } else {
         particles->m_spriteImageCount = 1;
         particles->m_blendImages = true;
     }
 
+    particles->m_hasTransparency = hasTransparency();
+
     if (m_colorTable)
         particles->m_colorTable = m_colorTable->getRenderImage();
     else
         particles->m_colorTable = nullptr;
 
+    if (!m_lights.isEmpty()) {
+        // Matches to QSSGRenderParticles lights
+        QVarLengthArray<QSSGRenderLight *, 4> lightNodes;
+        for (auto light : qAsConst(m_lights)) {
+            auto lightPrivate = QQuick3DObjectPrivate::get(light);
+            auto lightNode  = static_cast<QSSGRenderLight *>(lightPrivate->spatialNode);
+            lightNodes.append(lightNode);
+        }
+        particles->m_lights = lightNodes;
+    }
+
     particles->m_blendMode = mapBlendMode(m_blendMode);
-    particles->m_diffuseColor = color::sRGBToLinear(color());
     particles->m_billboard = m_billboard;
     particles->m_depthBias = perEmitter.emitter->depthBias();
+    particles->m_featureLevel = mapFeatureLevel(m_featureLevel);
+    particles->m_depthSorting = sortMode() == QQuick3DParticle::SortDistance;
 
     return particles;
 }
@@ -408,6 +487,25 @@ void QQuick3DParticleSpriteParticle::markNodesDirty()
         value.particleUpdateNode->m_nodeDirty = true;
 }
 
+void QQuick3DParticleSpriteParticle::updateFeatureLevel()
+{
+    FeatureLevel featureLevel = FeatureLevel::Simple;
+    if (m_lights.isEmpty()) {
+        if (m_colorTable)
+            featureLevel = FeatureLevel::Mapped;
+        if (m_spriteSequence)
+            featureLevel = FeatureLevel::Animated;
+    } else {
+        featureLevel = FeatureLevel::SimpleVLight;
+        if (m_colorTable)
+            featureLevel = FeatureLevel::MappedVLight;
+        if (m_spriteSequence)
+            featureLevel = FeatureLevel::AnimatedVLight;
+    }
+    if (featureLevel != m_featureLevel)
+        m_featureLevel = featureLevel;
+}
+
 void QQuick3DParticleSpriteParticle::componentComplete()
 {
     if (!system() && qobject_cast<QQuick3DParticleSystem *>(parentItem()))
@@ -421,10 +519,16 @@ void QQuick3DParticleSpriteParticle::reset()
     QQuick3DParticle::reset();
     deleteNodes();
     m_nextEmitterIndex = 0;
-    m_perEmitterData.clear();
     m_spriteParticleData.fill({});
 }
 
+void QQuick3DParticleSpriteParticle::commitParticles()
+{
+    markAllDirty();
+    update();
+    updateNodes();
+}
+
 int QQuick3DParticleSpriteParticle::nextCurrentIndex(const QQuick3DParticleEmitter *emitter)
 {
     if (!m_perEmitterData.contains(emitter)) {
@@ -450,10 +554,11 @@ void QQuick3DParticleSpriteParticle::setParticleData(int particleIndex,
                      const QVector3D &position,
                      const QVector3D &rotation,
                      const QVector4D &color,
-                     float size, float age)
+                     float size, float age,
+                     float animationFrame)
 {
     auto &dst = m_spriteParticleData[particleIndex];
-    dst = {position, rotation, color, size, age, dst.emitterIndex};
+    dst = {position, rotation, color, size, age, animationFrame, dst.emitterIndex};
 }
 
 void QQuick3DParticleSpriteParticle::updateParticleBuffer(const PerEmitterData &perEmitter, QSSGRenderGraphObject *spatialNode)
@@ -463,9 +568,10 @@ void QQuick3DParticleSpriteParticle::updateParticleBuffer(const PerEmitterData &
     if (!node)
         return;
     const int particleCount = perEmitter.particleCount;
-    if (node->m_particleBuffer.particleCount() != particleCount)
-        node->m_particleBuffer.resize(particleCount);
+    if (node->m_particleBuffer.particleCount() != particleCount || m_useAnimatedParticle)
+        node->m_particleBuffer.resize(particleCount, sizeof(QSSGParticleSimple));
 
+    m_useAnimatedParticle = false;
     char *dest = node->m_particleBuffer.pointer();
     const SpriteParticleData *src = particles.data();
     const int pps = node->m_particleBuffer.particlesPerSlice();
@@ -474,24 +580,127 @@ void QQuick3DParticleSpriteParticle::updateParticleBuffer(const PerEmitterData &
     const int emitterIndex = perEmitter.emitterIndex;
     int i = 0;
     QSSGBounds3 bounds;
-    for (int s = 0; s < slices; s++) {
-        QSSGParticle *dp = reinterpret_cast<QSSGParticle *>(dest);
-        for (int p = 0; p < pps && i < particleCount; ) {
-            if (src->emitterIndex == emitterIndex) {
-                if (src->size > 0.0f)
-                    bounds.include(src->position);
-                dp->position = src->position;
-                dp->rotation = src->rotation * float(M_PI / 180.0f);
-                dp->color = src->color;
-                dp->size = src->size * m_particleScale;
-                dp->age = src->age;
-                dp++;
-                p++;
-                i++;
+    const auto smode = sortMode();
+    if (smode == QQuick3DParticle::SortNewest || smode == QQuick3DParticle::SortOldest) {
+        int offset = m_currentIndex;
+        int step = (smode == QQuick3DParticle::SortNewest) ? -1 : 1;
+        int li = 0;
+        const auto sourceIndex = [&](int linearIndex, int offset, int wrap) -> int {
+            return (linearIndex + offset + wrap) % wrap;
+        };
+        for (int s = 0; s < slices; s++) {
+            QSSGParticleSimple *dp = reinterpret_cast<QSSGParticleSimple *>(dest);
+            for (int p = 0; p < pps && i < particleCount; ) {
+                const SpriteParticleData *data = src + sourceIndex(li * step, offset, m_maxAmount);
+                if (data->emitterIndex == emitterIndex) {
+                    if (data->size > 0.0f)
+                        bounds.include(data->position);
+                    dp->position = data->position;
+                    dp->rotation = data->rotation * float(M_PI / 180.0f);
+                    dp->color = data->color;
+                    dp->size = data->size * m_particleScale;
+                    dp->age = data->age;
+                    dp++;
+                    p++;
+                    i++;
+                }
+                li++;
             }
-            src++;
+            dest += ss;
+        }
+    } else {
+        for (int s = 0; s < slices; s++) {
+            QSSGParticleSimple *dp = reinterpret_cast<QSSGParticleSimple *>(dest);
+            for (int p = 0; p < pps && i < particleCount; ) {
+                if (src->emitterIndex == emitterIndex) {
+                    if (src->size > 0.0f)
+                        bounds.include(src->position);
+                    dp->position = src->position;
+                    dp->rotation = src->rotation * float(M_PI / 180.0f);
+                    dp->color = src->color;
+                    dp->size = src->size * m_particleScale;
+                    dp->age = src->age;
+                    dp++;
+                    p++;
+                    i++;
+                }
+                src++;
+            }
+            dest += ss;
+        }
+    }
+    node->m_particleBuffer.setBounds(bounds);
+}
+
+void QQuick3DParticleSpriteParticle::updateAnimatedParticleBuffer(const PerEmitterData &perEmitter, QSSGRenderGraphObject *spatialNode)
+{
+    const auto &particles = m_spriteParticleData;
+    QSSGRenderParticles *node = static_cast<QSSGRenderParticles *>(spatialNode);
+    if (!node)
+        return;
+    const int particleCount = perEmitter.particleCount;
+    if (node->m_particleBuffer.particleCount() != particleCount || !m_useAnimatedParticle)
+        node->m_particleBuffer.resize(particleCount, sizeof(QSSGParticleAnimated));
+
+    m_useAnimatedParticle = true;
+    char *dest = node->m_particleBuffer.pointer();
+    const SpriteParticleData *src = particles.data();
+    const int pps = node->m_particleBuffer.particlesPerSlice();
+    const int ss = node->m_particleBuffer.sliceStride();
+    const int slices = node->m_particleBuffer.sliceCount();
+    const int emitterIndex = perEmitter.emitterIndex;
+    int i = 0;
+    QSSGBounds3 bounds;
+    const auto smode = sortMode();
+    if (smode == QQuick3DParticle::SortNewest || smode == QQuick3DParticle::SortOldest) {
+        int offset = m_currentIndex;
+        int step = (smode == QQuick3DParticle::SortNewest) ? -1 : 1;
+        int li = 0;
+        const auto sourceIndex = [&](int linearIndex, int offset, int wrap) -> int {
+            return (linearIndex + offset + wrap) % wrap;
+        };
+        for (int s = 0; s < slices; s++) {
+            QSSGParticleAnimated *dp = reinterpret_cast<QSSGParticleAnimated *>(dest);
+            for (int p = 0; p < pps && i < particleCount; ) {
+                const SpriteParticleData *data = src + sourceIndex(li * step, offset, m_maxAmount);
+                if (data->emitterIndex == emitterIndex) {
+                    if (data->size > 0.0f)
+                        bounds.include(data->position);
+                    dp->position = data->position;
+                    dp->rotation = data->rotation * float(M_PI / 180.0f);
+                    dp->color = data->color;
+                    dp->size = data->size * m_particleScale;
+                    dp->age = data->age;
+                    dp->animationFrame = data->animationFrame;
+                    dp++;
+                    p++;
+                    i++;
+                }
+                li++;
+            }
+            dest += ss;
+        }
+    } else {
+        for (int s = 0; s < slices; s++) {
+            QSSGParticleAnimated *dp = reinterpret_cast<QSSGParticleAnimated *>(dest);
+            for (int p = 0; p < pps && i < particleCount; ) {
+                if (src->emitterIndex == emitterIndex) {
+                    if (src->size > 0.0f)
+                        bounds.include(src->position);
+                    dp->position = src->position;
+                    dp->rotation = src->rotation * float(M_PI / 180.0f);
+                    dp->color = src->color;
+                    dp->size = src->size * m_particleScale;
+                    dp->age = src->age;
+                    dp->animationFrame = src->animationFrame;
+                    dp++;
+                    p++;
+                    i++;
+                }
+                src++;
+            }
+            dest += ss;
         }
-        dest += ss;
     }
     node->m_particleBuffer.setBounds(bounds);
 }
@@ -508,4 +717,65 @@ void QQuick3DParticleSpriteParticle::updateSceneManager(QQuick3DSceneManager *sc
     }
 }
 
+// Lights
+void QQuick3DParticleSpriteParticle::onLightDestroyed(QObject *object)
+{
+    bool found = false;
+    for (int i = 0; i < m_lights.count(); ++i) {
+        if (m_lights[i] == object) {
+            m_lights.removeAt(i--);
+            found = true;
+        }
+    }
+    if (found) {
+        updateFeatureLevel();
+        markNodesDirty();
+    }
+}
+
+void QQuick3DParticleSpriteParticle::qmlAppendLight(QQmlListProperty<QQuick3DAbstractLight> *list, QQuick3DAbstractLight *light)
+{
+    if (!light)
+        return;
+
+    // Light must be id of an existing View3D light and not inline light element
+    if (light->parentItem()) {
+        QQuick3DParticleSpriteParticle *self = static_cast<QQuick3DParticleSpriteParticle *>(list->object);
+        self->m_lights.push_back(light);
+        self->updateFeatureLevel();
+        self->markNodesDirty();
+        // Make sure ligths are removed when destroyed
+        connect(light, &QQuick3DParticleSpriteParticle::destroyed, self, &QQuick3DParticleSpriteParticle::onLightDestroyed);
+    }
+}
+
+QQuick3DAbstractLight *QQuick3DParticleSpriteParticle::qmlLightAt(QQmlListProperty<QQuick3DAbstractLight> *list, qsizetype index)
+{
+    QQuick3DParticleSpriteParticle *self = static_cast<QQuick3DParticleSpriteParticle *>(list->object);
+    if (index >= self->m_lights.size()) {
+        qWarning("The index exceeds the range of valid light targets.");
+        return nullptr;
+    }
+    return self->m_lights.at(index);
+}
+
+qsizetype QQuick3DParticleSpriteParticle::qmlLightsCount(QQmlListProperty<QQuick3DAbstractLight> *list)
+{
+    QQuick3DParticleSpriteParticle *self = static_cast<QQuick3DParticleSpriteParticle *>(list->object);
+    return self->m_lights.count();
+}
+
+void QQuick3DParticleSpriteParticle::qmlClearLights(QQmlListProperty<QQuick3DAbstractLight> *list)
+{
+    QQuick3DParticleSpriteParticle *self = static_cast<QQuick3DParticleSpriteParticle *>(list->object);
+    for (const auto &light : qAsConst(self->m_lights)) {
+        if (light->parentItem() == nullptr)
+            QQuick3DObjectPrivate::get(light)->derefSceneManager();
+        light->disconnect(self, SLOT(onLightDestroyed(QObject*)));
+    }
+    self->m_lights.clear();
+    self->updateFeatureLevel();
+    self->markNodesDirty();
+}
+
 QT_END_NAMESPACE
